@@ -4,9 +4,11 @@ understand the patterns and relationships of geographic features.
 This is mainly inspired by turf.js.
 link: http://turfjs.org/
 """
-
+import concurrent.futures
+from functools import partial
 from math import asin, atan2, cos, degrees, log, pi, pow, radians, sin, sqrt, tan
-from typing import Optional, Union
+from multiprocessing import Manager
+from typing import List, Optional, Union
 
 from geojson import (
     Feature,
@@ -33,6 +35,7 @@ from turfpy.helper import (
 from turfpy.meta import (
     coord_each,
     feature_each,
+    geom_each,
     geom_reduce,
     segment_each,
     segment_reduce,
@@ -74,8 +77,7 @@ def bearing(start: Feature, end: Feature, final=False) -> float:
 
 
 def _calculate_final_bearing(start, end) -> float:
-    """ #TODO: Add description
-    """
+    """#TODO: Add description"""
     bear = bearing(end, start)
     bear = (bear + 180) % 360
     return bear
@@ -144,7 +146,7 @@ def area(
     This function calculates the area of the Geojson object given as input.
 
     :param geojson: Geojson object for which area is to be found.
-    :return: area for the given Geojson object.
+    :return: area for the given Geojson object in square meters.
 
     Example:
 
@@ -169,9 +171,11 @@ def area(
 def bbox(geojson):
     """
     This function is used to generate bounding box coordinates for given geojson.
+
     :param geojson: Geojson object for which bounding box is to be found.
     :return: bounding box for the given Geojson object.
-    Example :-
+
+    Example :
 
     >>> from turfpy.measurement import bbox
     >>> from geojson import Polygon
@@ -206,10 +210,13 @@ def bbox(geojson):
 def bbox_polygon(bbox: list, properties: dict = {}) -> Feature:
     """
     To generate a Polygon Feature for the bounding box generated using bbox.
+
     :param bbox: bounding box generated for a geojson.
     :param properties: properties to be added to the returned feature.
     :return: polygon for the given bounding box coordinates.
-    Example :-
+
+    Example :
+
     >>> from turfpy.measurement import bbox_polygon, bbox
     >>> from geojson import Polygon
 
@@ -257,10 +264,13 @@ def center(geojson, properties: Optional[dict] = None) -> Feature:
     """
     Takes a Feature or FeatureCollection and returns the absolute center point of all
     features.
+
     :param geojson: GeoJSON for which centered to be calculated.
     :param properties: Optional parameters to be set to the generated feature.
     :return: Point feature for the center.
-    Example :-
+
+    Example :
+
     >>> from turfpy.measurement import center
     >>> from geojson import Feature, FeatureCollection, Point
 
@@ -303,9 +313,12 @@ def envelope(geojson) -> Feature:
     """
     Takes any number of features and returns a rectangular Polygon that encompasses all
     vertices.
+
     :param geojson: geojson input features for which envelope to be generated.
     :return: returns envelope i.e bounding box polygon.
-    Example :-
+
+    Example :
+
     >>> from turfpy.measurement import envelope
     >>> from geojson import Feature, FeatureCollection, Point
 
@@ -326,10 +339,13 @@ def envelope(geojson) -> Feature:
 def length(geojson, units: str = "km"):
     """
     Takes a geojson and measures its length in the specified units.
+
     :param geojson: geojson for which the length is to be determined.
     :param units: units in which length is to be returned.
     :return: length of the geojson in specified units.
-    Example:-
+
+    Example:
+
     >>> from turfpy.measurement import length
     >>> from geojson import LineString
     >>> ls = LineString([(115, -32), (131, -22), (143, -25), (150, -34)])
@@ -350,19 +366,23 @@ def length(geojson, units: str = "km"):
 # ----------- Destination --------------#
 
 
-def destination(origin: Feature, distance, bearing, options: dict = {}) -> Feature:
+def destination(
+    origin: Union[Feature, Point], distance, bearing, options: dict = {}
+) -> Feature:
     """
     Takes a Point and calculates the location of a destination point given a distance in
-    degrees, radians, miles,
-    or kilometers and bearing in degrees.
+    degrees, radians, miles, or kilometers and bearing in degrees.
+
     :param origin: Start point.
     :param distance: distance upto which the destination is from origin.
     :param bearing: Direction in which is the destination is from origin.
     :param options: Option like units of distance and properties to be passed to
         destination point feature, value
-    for units are 'mi', 'km', 'deg' and 'rad'.
+        for units are 'mi', 'km', 'deg' and 'rad'.
     :return: Feature: destination point in at the given distance and given direction.
-    Example :-
+
+    Example:
+
     >>> from turfpy.measurement import destination
     >>> from geojson import Point, Feature
     >>> origin = Feature(geometry=Point([-75.343, 39.984]))
@@ -371,7 +391,7 @@ def destination(origin: Feature, distance, bearing, options: dict = {}) -> Featu
     >>> options = {'units': 'mi'}
     >>> destination(origin,distance,bearing,options)
     """
-    coordinates1 = origin["geometry"]["coordinates"]
+    coordinates1 = get_coord(origin)
     longitude1 = radians(float(coordinates1[0]))
     latitude1 = radians(float(coordinates1[1]))
     bearingRad = radians(float(bearing))
@@ -407,10 +427,13 @@ def destination(origin: Feature, distance, bearing, options: dict = {}) -> Featu
 def centroid(geojson, properties: dict = None) -> Feature:
     """
     Takes one or more features and calculates the centroid using the mean of all vertices.
+
     :param geojson: Input features
     :param properties: Properties to be set to the output Feature point
     :return: Feature: Point feature which is the centroid of the given features
-    Example:-
+
+    Example:
+
     >>> from turfpy.measurement import centroid
     >>> from geojson import Polygon
     >>> polygon = Polygon([((-81, 41), (-88, 36), (-84, 31), (-80, 33), (-77, 39),
@@ -419,18 +442,18 @@ def centroid(geojson, properties: dict = None) -> Feature:
     """
     x_sum = 0
     y_sum = 0
-    len = 0
+    length = 0
 
     def _callback_coord_each(
         coord, coord_index, feature_index, multi_feature_index, geometry_index
     ):
-        nonlocal x_sum, y_sum, len
+        nonlocal x_sum, y_sum, length
         x_sum += coord[0]
         y_sum += coord[1]
-        len += 1
+        length += 1
 
     coord_each(geojson, _callback_coord_each)
-    point = Point((x_sum / len, y_sum / len))
+    point = Point((x_sum / length, y_sum / length))
     return Feature(geometry=point, properties=properties if properties else {})
 
 
@@ -442,11 +465,14 @@ def centroid(geojson, properties: dict = None) -> Feature:
 def along(line: Feature, dist, unit: str = "km") -> Feature:
     """
     This function is used identify a Point at a specified distance along a LineString.
+
     :param line: LineString on which the point to be identified
     :param dist: Distance from the start of the LineString
     :param unit: unit of distance
     :return: Feature : Point at the distance on the LineString passed
-    Example :-
+
+    Example :
+
     >>> from turfpy.measurement import along
     >>> from geojson import LineString, Feature
     >>> ls = Feature(geometry=LineString([(-83, 30), (-84, 36), (-78, 41)]))
@@ -500,10 +526,13 @@ def along(line: Feature, dist, unit: str = "km") -> Feature:
 def midpoint(point1: Feature, point2: Feature) -> Feature:
     """
     This function is used to get midpoint between any the two points.
+
     :param point1: First point.
     :param point2: Second point.
     :return: Feature: Point which is the midpoint of the two points given as input.
-    Example:-
+
+    Example:
+
     >>> from turfpy.measurement import midpoint
     >>> from geojson import Point, Feature
     >>> point1 = Feature(geometry=Point((144.834823, -37.771257)))
@@ -525,8 +554,8 @@ def midpoint(point1: Feature, point2: Feature) -> Feature:
 def nearest_point(target_point: Feature, points: FeatureCollection) -> Feature:
     """
     Takes a reference Point Feature and FeatureCollection of point features and returns
-    the point from the
-    FeatureCollection closest to the reference Point Feature.
+    the point from the FeatureCollection closest to the reference Point Feature.
+
     :param target_point: Feature Point of reference.
     :param points: FeatureCollection of points.
     :return: a Point Feature from the FeatureCollection which is closest to the reference
@@ -558,6 +587,7 @@ def nearest_point(target_point: Feature, points: FeatureCollection) -> Feature:
         if float(distance_to_point) < min_dist:
             best_feature_index = feature_index
             min_dist = distance_to_point
+        return True
 
     feature_each(points, _callback_feature_each)
 
@@ -576,9 +606,12 @@ def point_on_feature(geojson) -> Feature:
     """
     Takes a Feature or FeatureCollection and returns a Point guaranteed to be on the
     surface of the feature.
+
     :param geojson: Feature or FeatureCollection on which the Point is to be found.
     :return: Feature point which on the provided feature.
-    Example:-
+
+    Example:
+
     >>> from turfpy.measurement import point_on_feature
     >>> from geojson import  Polygon, Feature
     >>> point = Polygon([((116, -36), (131, -32), (146, -43), (155, -25), (133, -9),
@@ -683,13 +716,15 @@ def boolean_point_in_polygon(point, polygon, ignore_boundary=False):
     """
     Takes a Point or a Point Feature and Polygon or Polygon Feature as input and returns
     True if Point is in given Feature.
+
     :param point: Point or Point Feature.
     :param polygon: Polygon or Polygon Feature.
     :param ignore_boundary: [Optional] default value is False, specify whether to exclude
-        boundary of the given polygon
-    or not.
+        boundary of the given polygon or not.
     :return: True if the given Point is in Polygons else False
-    Example:-
+
+    Example:
+
     >>> from turfpy.measurement import boolean_point_in_polygon
     >>> from geojson import Point, MultiPolygon, Feature
     >>> point = Feature(geometry=Point((-77, 44)))
@@ -772,7 +807,11 @@ def explode(geojson):
 
         def _callback_feature_each(feature, feature_index):
             def _callback_coord_each(
-                coord, coord_index, feature_index, multi_feature_index, geometry_index,
+                coord,
+                coord_index,
+                feature_index,
+                multi_feature_index,
+                geometry_index,
             ):
                 nonlocal points
                 point = Point(coord)
@@ -784,7 +823,11 @@ def explode(geojson):
     else:
 
         def _callback_coord_each(
-            coord, coord_index, feature_index, multi_feature_index, geometry_index,
+            coord,
+            coord_index,
+            feature_index,
+            multi_feature_index,
+            geometry_index,
         ):
             nonlocal points, geojson
             point = Point(coord)
@@ -802,10 +845,13 @@ def explode(geojson):
 def polygon_tangents(point, polygon):
     """
     Finds the tangents of a (Multi)Polygon from a Point.
+
     :param point: Point or Point Feature.
     :param polygon: (Multi)Polygon or (Multi)Polygon Feature.
     :return: FeatureCollection of two tangent Point Feature.
-    Example:-
+
+    Example:
+
     >>> from turfpy.measurement import polygon_tangents
     >>> from geojson import Polygon, Point, Feature
     >>> point = Feature(geometry=Point((61, 5)))
@@ -838,7 +884,9 @@ def polygon_tangents(point, polygon):
                 ltan = poly_coords[0][nearest_pt_index]
 
         eprev = _is_left(
-            poly_coords[0][0], poly_coords[0][len(poly_coords[0]) - 1], point_coords,
+            poly_coords[0][0],
+            poly_coords[0][len(poly_coords[0]) - 1],
+            point_coords,
         )
         out = process_polygon(poly_coords[0], point_coords, eprev, enext, rtan, ltan)
         rtan = out[0]
@@ -915,13 +963,16 @@ def _is_left(point1, point2, point3):
 def point_to_line_distance(point: Feature, line: Feature, units="km", method="geodesic"):
     """
     Returns the minimum distance between a Point and any segment of the LineString.
+
     :param point: Point Feature from which distance to be measured.
     :param line: Point LineString from which distance to be measured.
     :param units: units for distance 'km', 'm', 'mi, 'ft', 'in', 'deg', 'cen', 'rad',
         'naut', 'yd'
     :param method: Method which is used to calculate, values can be 'geodesic' or 'planar'
     :return: Approximate distance between the LineString and Point
-    Example:-
+
+    Example:
+
     >>> from turfpy.measurement import point_to_line_distance
     >>> from geojson import LineString, Point, Feature
     >>> point = Feature(geometry=Point((0, 0)))
@@ -1045,8 +1096,7 @@ def rhumb_bearing(start, end, final=False):
 
 
 def calculate_rhumb_bearing(fro, to):
-    """ #TODO: Add description
-    """
+    """#TODO: Add description"""
     phi1 = radians(fro[1])
     phi2 = radians(to[1])
     delta_lambda = radians(to[0] - fro[0])
@@ -1068,10 +1118,11 @@ def calculate_rhumb_bearing(fro, to):
 # ------------ rhumb destination -----------#
 
 
-def rhumb_destination(origin, distance, bearing, options) -> Feature:
+def rhumb_destination(origin, distance, bearing, options: dict = {}) -> Feature:
     """
     Returns the destination Point having travelled the given distance along a Rhumb line
     from the origin Point with the (varant) given bearing.
+
     :param origin: Starting Point
     :param distance: Distance from the starting point
     :param bearing: Varant bearing angle ranging from -180 to 180 degrees from north
@@ -1092,13 +1143,14 @@ def rhumb_destination(origin, distance, bearing, options) -> Feature:
     'properties': {"marker-color": "F00"}})
     """
     was_negative_distance = distance < 0
-    distance_in_meters = convert_length(abs(distance), options.get("units", ""), "m")
+    distance_in_meters = convert_length(abs(distance), options.get("units", "km"), "m")
     if was_negative_distance:
         distance_in_meters = -1 * (abs(distance_in_meters))
     coords = get_coord(origin)
     destination_point = _calculate_rhumb_destination(coords, distance_in_meters, bearing)
     return Feature(
-        geometry=Point(destination_point), properties=options.get("properties", ""),
+        geometry=Point(destination_point),
+        properties=options.get("properties", ""),
     )
 
 
@@ -1148,7 +1200,9 @@ def rhumb_distance(start, to, units="km"):
     :param units: Units in which distance to be calculated, values can be 'deg', 'rad',
         'mi', 'km'
     :return: Distance calculated from provided start to end Point.
-    Example:-
+
+    Example:
+
     >>> from turfpy.measurement import rhumb_distance
     >>> from geojson import Point, Feature
     >>> start = Feature(geometry=Point((-75.343, 39.984)))
@@ -1243,3 +1297,65 @@ def square(bbox: list):
 
 
 # -------------------------------#
+
+
+def points_within_polygon(
+    points: Union[Feature, FeatureCollection],
+    polygons: Union[Feature, FeatureCollection],
+    chunk_size: int = 1,
+) -> FeatureCollection:
+    """Find Point(s) that fall within (Multi)Polygon(s).
+
+    This function takes two inputs GeoJSON Feature :class:`geojson.Point` or
+    :class:`geojson.FeatureCollection` of Points and GeoJSON Feature
+    :class:`geojson.Polygon` or Feature :class:`geojson.MultiPolygon` or
+    FeatureCollection of :class:`geojson.Polygon` or Feature
+    :class:`geojson.MultiPolygon`. and returns all points with in in those
+    Polygon(s) or (Multi)Polygon(s).
+
+    :param points: A single GeoJSON ``Point`` feature or FeatureCollection of Points.
+    :param polygons: A Single GeoJSON Polygon/MultiPolygon or FeatureCollection of
+        Polygons/MultiPolygons.
+    :param chunk_size: Number of chunks each process to handle. The default value is
+            1, for a large number of features please use `chunk_size` greater than 1
+            to get better results in terms of performance.
+    :return: A :class:`geojson.FeatureCollection` of Points.
+    """
+    if not points:
+        raise Exception("Points cannot be empty")
+
+    if points["type"] == "Point":
+        points = FeatureCollection([Feature(geometry=points)])
+
+    if points["type"] == "Feature":
+        points = FeatureCollection([points])
+
+    manager = Manager()
+    results: List[dict] = manager.list()
+
+    part_func = partial(
+        check_each_point,
+        polygons=polygons,
+        results=results,
+    )
+
+    with concurrent.futures.ProcessPoolExecutor() as executor:
+        for _ in executor.map(part_func, points["features"], chunksize=chunk_size):
+            pass
+
+    return FeatureCollection(list(results))
+
+
+def check_each_point(point, polygons, results):
+    def __callback_geom_each(
+        current_geometry, feature_index, feature_properties, feature_bbox, feature_id
+    ):
+        contained = False
+        if boolean_point_in_polygon(point, current_geometry):
+            contained = True
+
+        if contained:
+            nonlocal results
+            results.append(point)
+
+    geom_each(polygons, __callback_geom_each)
